/*
	System: Structured text retrieval tool sgrep.
	Module: pmatch.c
	Author: Pekka Kilpelinen & Jani Jaakkola
	Description: Pattern matching using Aho-Corasick automate (ACsearch() )
	Version history: Original version February 1995 by JJ & PK
	Copyright: University of Helsinki, Dept. of Computer Science
		   Distributed under GNU General Public Lisence
		   See file COPYING for details
*/

/* NOTE: Aho-Corasick automate can only take constant patterns. There is
         no wild card expansions and it's always case sensitive. Maybe
         something should be done about this.
*/

/*
 * NOTE: This module is used both by normal query engine and indexing
 * engine. query engine accesses this module through search().
 * Interface to indexer is a bit awkward: index.c calls index() function
 * to start scanning files and this module calls add_region_index() 
 * to add the newly found regions to index. This is a bit awkward and
 * possibly also somewhat (little?) inefficient.
 */


#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define SGREP_LIBRARY
#include "sgrep.h"

/* These come from the XML-specs */
#define NameChars ("-a-zA-Z0-9._:")
#define NameStartChars ("a-zA-Z_:")

#define WordChars ("a-zA-Z")

/*
 * Private data structures
 */

/* Scanbuffer is for reading files */
struct ScanBuffer {
    SgrepData *sgrep;
    FileList *file_list;
    int len;
    int file_num;
    int old_file_num;
    int last_file;
    int region_start;
    const unsigned char *map;
    int map_size;
};


/* OutputList, ACState and ACScanner are for Aho-Corasick automate */
struct OutputList {
	struct PHRASE_NODE *phrase;
	struct OutputList *next;
};

struct ACState {
    struct ACState *gotos[256];
    struct ACState *fail;
    struct ACState *next; /* queue needed when creating fail function */
    struct OutputList *output_list;
#ifdef DEBUG
    int state_num;
#endif
};

struct ACScanner {
    SgrepData *sgrep;
    struct PHRASE_NODE *phrase_list; /* Points to this scanners phrase list */
    struct ACState *root_state; /* Points to root state */
    struct ACState *s;          /* Points to current state */
    int ignore_case;
} AC_scanner; /* THE AC_scanner, since only one is needed */

/*
 * datatypes for the SGML-scanner
 */

/*
 * Implementation of a simple SGML scanner starts here
 * (simple finite automaton)
 */
/* FIXME: index.c has a different name for same constant */
#define MAX_TERM_SIZE 256
enum SGMLState { SGML_PCDATA, SGML_WORD, 
		 SGML_ENTITY_OPEN,SGML_ENTITY,SGML_PCDATA_ENTITY,
		 SGML_WORD_ENTITY,
		 SGML_CHARACTER_REFERENCE_OPEN,
		 SGML_DECIMAL_CHARACTER_REFERENCE,
		 SGML_HEX_CHARACTER_REFERENCE,
		 SGML_CHARACTER_REFERENCE_CLOSE,		 
		 SGML_STAGO, SGML_GI, SGML_W_ATTNAME,
                 SGML_ATTNAME,
                 SGML_W_ATTEQUAL, SGML_W_ATTVALUE, 
		 SGML_ATTVALUE, SGML_ATTVALUE_DQUOTED, 
		 SGML_ATTVALUE_SQUOTED,
		 SGML_STAGC, SGML_END_TAG, SGML_W_ETAGC,
		 SGML_DECLARATION_START,SGML_DOCTYPE_DECLARATION,
		 SGML_COMMENT_START, SGML_COMMENT, SGML_COMMENT_WORD,
		 SGML_COMMENT_END1, SGML_COMMENT_END2,
		 SGML_PI_START, SGML_PI, SGML_PI_END,
		 SGML_MARKED_SECTION_START, SGML_MARKED_SECTION_START2,
		 SGML_CDATA_MARKED_SECTION, SGML_CDATA_MARKED_SECTION_WORD,
		 SGML_CDATA_MARKED_SECTION_END1,
		 SGML_CDATA_MARKED_SECTION_END2,
		 SGML_DOCTYPE, SGML_DOCTYPE_EXTERNAL,
		 SGML_DOCTYPE_PUBLIC, SGML_DOCTYPE_SYSTEM,
		 SGML_WAITING_LITERAL,
		 SGML_LITERAL_START,SGML_LITERAL_SQUOTED,SGML_LITERAL_DQUOTED,
		 SGML_DOCTYPE_PUBLIC_ID_START, SGML_DOCTYPE_PUBLIC_ID,
		 SGML_DOCTYPE_SYSTEM_ID_START, SGML_DOCTYPE_SYSTEM_ID,
		 SGML_DOCTYPE_INTERNAL_START,
		 SGML_DOCTYPE_INTERNAL,
		 SGML_INTERNAL_DECLARATION_START1,
		 SGML_INTERNAL_DECLARATION_START2,
		 SGML_PEREFERENCE,
		 SGML_INTERNAL_DECLARATION_NAME,
		 SGML_ENTITY_DECLARATION,
		 SGML_ENTITY_DECLARATION_NAME,
		 SGML_ENTITY_DEFINITION,
		 SGML_ENTITY_DEFINITION_TYPE,
		 SGML_LITERAL_ENTITY,
		 SGML_GENERAL_ENTITY_DEFINITION_DQUOTED,
		 SGML_GENERAL_ENTITY_DEFINITION_SQUOTED,
		 SGML_ENTITY_DEFINITION_PUBLIC_ID,
		 SGML_WAITING_ENTITY_DEFINITION_SYSTEM_ID,
		 SGML_ENTITY_DEFINITION_SYSTEM_ID,
		 SGML_ENTITY_DEFINITION_NDATA,
		 SGML_ENTITY_DEFINITION_NDATA2,
		 SGML_ENTITY_DEFINITION_NDATA_NAME,
		 SGML_ENTITY_DEFINITION_END,
		 SGML_ELEMENT_TYPE_DECLARATION,
		 SGML_ATTLIST_DECLARATION,
		 SGML_NOTATION_DECLARATION,
		 SGML_DOCTYPE_END,
		 SGML_RESERVED_WORD
                };

typedef struct ElementStackStruct {
    char *gi;
    int start;
    int end;
    struct ElementStackStruct *prev;
} ElementStack;

typedef struct SGMLScannerStruct {
    SgrepData *sgrep;
    struct ScanBuffer *scan_buffer;

    /* Scanner "configuration" */
    ScannerType type;
    char *name_start_chars;
    char *name_chars;
    char *word_chars;
    int ignore_case;
    int include_system_entities;

    /* Maintaining the element stack */
    int maintain_element_stack;
    ElementStack *top;
    RegionList *element_list;
    
    /* Scanner state */
    int parse_errors;
    struct PHRASE_NODE *phrase_list;
    int wordi;
    int words;
    int word_end;
    unsigned char word[MAX_TERM_SIZE];
    int sgii;
    int sgis;
    unsigned char sgi[MAX_TERM_SIZE];
    int egii;
    int egis;
    unsigned char egi[MAX_TERM_SIZE];
    int anames;
    int anamei;
    unsigned char aname[MAX_TERM_SIZE];
    int avali;
    int avals;
    unsigned char aval[MAX_TERM_SIZE];
    int comments;
    int comment_wordi;
    int comment_words;
    unsigned char comment_word[MAX_TERM_SIZE];

    int markeds;
    
    int doctypes;
    int doctype_declarations;
    int internal_declarations;

    int entity_has_systemid;
    int entity_is_ndata;
    int entitys; /* Start of entity reference */
    int character_reference;

    int namei;
    unsigned char name[MAX_TERM_SIZE];

    int literals;
    int literali;
    unsigned char literal[MAX_TERM_SIZE];

    int publici;
    int systemi;

    int pis;
    int pii;
    unsigned char pi[MAX_TERM_SIZE];

    enum SGMLState state;
    
    /* This isn't a real stack of context free language parser;
     * it's just a trick for some state reuse */
#define SGML_SCANNER_STACK_SIZE 10
    enum SGMLState state_stack[SGML_SCANNER_STACK_SIZE];
    int state_stack_ptr;

    void (*entry)(struct SGMLScannerStruct *state,
		  const char *str, int start, int end);
    void *data;

    int failed;
} SGMLScanner;

/*
 * Private prototypes
 */
void new_output( SgrepData *sgrep, struct ACState *s, 
			struct PHRASE_NODE *pn);
void ACsearch(struct ACScanner *scanner, const unsigned char *buf, 
	      int len, int start);
void enter(SgrepData *sgrep, struct PHRASE_NODE *pn, 
		  struct ACState *root_state, int ignore_case);
void create_fail(SgrepData *sgrep, struct ACState *root_state);
void create_goto(SgrepData *sgrep,struct PHRASE_NODE *phrase_list, 
			struct ACState *root_state,
			int ignore_case);
void pop_elements_to(SGMLScanner *state,
			    struct ElementStackStruct *p);



#ifdef DEBUG
void show_states(int depth,struct State_Data *s);
void show_ACtree();
void show_outputs(struct State_Data *s);
#endif



void print_scanner_help() {
    printf("      %-12s%s\n","sgml","use SGML scanner");
    printf("      %-12s%s\n","html","use HTML scanner (currently same as sgml scanner)");
    printf("      %-12s%s\n","xml","use XML scanner");
    printf("      %-12s%s\n","sgml-debug","show recognized SGML tokens");
    printf("      %-12s%s\n","include-entities","  automatically include system entities");
}

int set_scanner_option(SgrepData *sgrep,const char *arg) {
    if ( strcmp(arg,"sgml")==0 || 
	 strcmp(arg,"html")==0) {
	sgrep->scanner_type=SGML_SCANNER;
    } else if (strcmp(arg,"xml")==0) {
	sgrep->scanner_type=XML_SCANNER;
    } else if (strcmp(arg,"sgml-debug")==0) {
	sgrep->sgml_debug=1;
    } else if (strcmp(arg,"include-entities")==0) {
	sgrep->include_system_entities=1;
    } else {
	sgrep_error(sgrep,"Unknown scanner argument '%s'\n",arg);
	return SGREP_ERROR;
    }
    return SGREP_OK;
}


/*
 * Parses a given character list returning an array containing accepted
 * characters
 */
char *expand_character_list(SgrepData *sgrep,const char *l) {
    char *a;
    int i,j;
    const unsigned char *list=(unsigned char *)l;

    a=(char *)sgrep_calloc(256,sizeof(char));
    for(i=0;list[i];i++) {
	if (list[i]=='-' && i>0) {
	    /* A region */
	    if (list[i+1]==0) {
		sgrep_error(sgrep,"Character list '%s' contains a region with no endpoint",
			list);
	    } else {
		    for(j=list[i-1];j<=list[i+1];j++) {
			a[j]=1;
			}
	    }
	    i++;
	} else {
	    a[list[i]]=1;
	}
    }
    return a;
}

/*
 * Creates and initializes a new scanner buffer struct with given
 * input files
 */
struct ScanBuffer *new_scan_buffer(SgrepData *sgrep,FileList *files) {
    struct ScanBuffer *sc;
    sc=sgrep_new(struct ScanBuffer);
    sc->sgrep=sgrep;
    sc->file_list=files;
    sc->len=0;
    sc->file_num=0;
    sc->old_file_num=-1;
    sc->last_file=-1; /* As many files there is in list */
    sc->region_start=0;
    sc->map=NULL;
    sc->map_size=0;
    return sc;
}

struct ScanBuffer *reset_scan_buffer(struct ScanBuffer *sc, 
				     int f_file, int l_file) {
    sc->file_num=f_file;
    sc->last_file=l_file;
    sc->region_start=flist_start(sc->file_list,f_file);
    return sc;
}


void delete_scan_buffer(struct ScanBuffer *b) {
    SgrepData *sgrep=b->sgrep;
    sgrep_free(b);
}

/*
 * Fills the scanner input buffer and sets len and file_num respectively
 * if all files have been processed returns 0
 * otherwise returns number of bytes available in buffer
 * NEW: uses map_file() instead of read()
 */
int next_scan_buffer(struct ScanBuffer *sb)
{
    SGREPDATA(sb);
    if (sb->map && sb->len==sb->map_size) {
	sb->file_num++;
    }
    /* Skip zero length files */
    while(sb->file_num<flist_files(sb->file_list) &&
	  flist_length(sb->file_list,sb->file_num)==0) {
	sb->file_num++;
    }
    if (sb->old_file_num!=sb->file_num && sb->map) {
	unmap_file(sgrep,(void *)sb->map,sb->map_size);
	sb->map=NULL;
	sb->map_size=0;
    }
    if ( (sb->last_file==-1 && sb->file_num>=flist_files(sb->file_list)) ||
	 (sb->last_file>=0 && sb->file_num>sb->last_file) ) {
	/* All files scanned */
	return 0;
    }
    if (!sb->map) {
	void *map;
	sb->map_size=map_file(sgrep,flist_name(sb->file_list,sb->file_num),
			      &map);
	sb->map=(const unsigned char*)map;
    }
    if (sb->map==NULL) {
	sgrep_error(sgrep,"Failed to scan file '%s'\n",
		    flist_name(sb->file_list,sb->file_num));
	return SGREP_ERROR;
    }
    sb->old_file_num=sb->file_num;
    if (sb->map_size!=flist_length(sb->file_list,sb->file_num)) {
	sgrep_error(sgrep,"Size of file '%s' has changed\n",
		    flist_name(sb->file_list,sb->file_num));
    }
    sb->region_start+=sb->len;
    sb->len=sb->map_size;
    return sb->len;
}
		
/*
 * Gives and inits a new state to the automate.
 */
/* If root_state is NULL
 * it inits the root_state 
 */
struct ACState *new_state(SgrepData *sgrep)
{
	int i;
	struct ACState *s;
#ifdef DEBUG
	static int snum;
#endif
	s=(struct ACState *)sgrep_malloc( sizeof(struct ACState));
	for (i=0;i<256;i++) s->gotos[i]=NULL;
	s->output_list=NULL;
	s->next=NULL;
	s->fail=NULL;
#ifdef DEBUG
	snum++;
	s->state_num=snum;
#endif
	return s;
}

/*
 * Enters a new output link to a state 
 */
void new_output(SgrepData *sgrep, struct ACState *s, struct PHRASE_NODE *pn)
{
	struct OutputList **op;

	op=&s->output_list;
	while (*op!=NULL) op=& (*op)->next;
	*op=sgrep_new(struct OutputList);
	(*op)->next=NULL;
	(*op)->phrase=pn;
}

/*
 * Enters a new phrase to automate given with root_state
 */
void enter(SgrepData *sgrep, struct PHRASE_NODE *pn, 
	   struct ACState *root_state,
	   int ignore_case)
{
	struct ACState *state=root_state;
	size_t j;
	unsigned char pch;

#ifdef DEBUG
	printf("enter %s",str);
#endif
	assert(pn->phrase->s[0]=='n');
	j=1;
	pch=pn->phrase->s[j];
	if (ignore_case) pch=toupper(pch);
	while ( state->gotos[pch]!=NULL && j<pn->phrase->length )
	{
		state=state->gotos[pch];
		j++;
		pch=pn->phrase->s[j];
		if (ignore_case) pch=toupper(pch);
	}
	
	while( j<pn->phrase->length )
	{
		state->gotos[pch]=new_state(sgrep);
		state=state->gotos[pch];
		j++;
		pch=pn->phrase->s[j];
		if (ignore_case) pch=toupper(pch);
	}
	new_output(sgrep,state,pn);
#ifdef DEBUG
	printf(" done\n");
#endif
}
	
/*
 * The creation of the AC goto function using the enter function 
 * and the phrase list. 
 * The automate is spesified with root_state
 */
void create_goto(SgrepData *sgrep, struct PHRASE_NODE *phrase_list,
		 struct ACState *root_state,
		 int ignore_case)
{
	struct PHRASE_NODE *pn;

	for(pn=phrase_list;pn!=NULL;pn=pn->next) {
	    if (pn->phrase->s[0]=='n') {
		/* Add only AC phrases to automate */
#if DEBUG
		fprintf(stderr,"AC phrase:%s\n",lpn->phrase->s);
#endif		
		enter(sgrep,pn,root_state,ignore_case);
	    }
	}
}

/*
 * The creation of the AC fail function and the final output function 
 * The automate to use is given with root_state
 */
void create_fail(SgrepData *sgrep,struct ACState *root_state) 
{
	int i;
	struct ACState *s,*r,*state;
	struct ACState *first=NULL;
	struct ACState *last=NULL;	
	struct OutputList *op;
	
#ifdef DEBUG
	printf("Create fail :");
#endif	
	for (i=0;i<256;i++)
	{
		if ( (s=root_state->gotos[i]) !=root_state )
		{		
			if (first==NULL) first=s;
			if (last!=NULL) last->next=s;
			last=s;
			last->next=NULL;
			s->fail=root_state;
		}
	}
#ifdef DEBUG
	printf(" root done");
#endif
	while (first!=NULL)
	{
		r=first;
		first=first->next;
		for (i=0;i<256;i++) if ( r->gotos[i]!=NULL )
		{
			s=r->gotos[i];
			last->next=s;
			last=s;
			last->next=NULL;
			if (first==NULL) first=last;
			state=r->fail;
			while (state->gotos[i]==NULL) state=state->fail;
			s->fail=state->gotos[i];
			for (op=s->fail->output_list;
			     op!=NULL;
			     op=op->next) {
			    assert(op->phrase!=NULL);
			    new_output(sgrep,s,op->phrase);
			}
		}
	}
#ifdef DEBUG
	printf(", all done\n");
#endif
}

/* Phrases list points to list of phrases to be
 * matched. ifs points to names of input files, and lf is the number of
 * input files.
 */
struct ACScanner *init_AC_search(SgrepData *sgrep,
				 struct PHRASE_NODE *phrase_list) {
    int i;
    struct ACScanner *sc;

    sc=sgrep_new(struct ACScanner);
    sc->sgrep=sgrep;
    sc->root_state=new_state(sgrep);
    sc->phrase_list=phrase_list;
    sc->s=sc->root_state;
    sc->ignore_case=sgrep->ignore_case;
    create_goto(sgrep,phrase_list,sc->root_state,sc->ignore_case);
    /* there isn't any fail links from root state */
    for (i=0;i<256;i++) {
	if (sc->root_state->gotos[i]==NULL) 
	    sc->root_state->gotos[i]=sc->root_state;
    }
    create_fail(sgrep,sc->root_state);
    return sc;
}

void delete_AC_state(SgrepData *sgrep,struct ACState *as) {
    int i;
    for(i=0;i<256;i++) {
	if (as->gotos[i] && as->gotos[i]!=as) {
	    delete_AC_state(sgrep,as->gotos[i]);
	}
	while(as->output_list) {
	    struct OutputList *ol=as->output_list;
	    as->output_list=as->output_list->next;
	    sgrep_free(ol);
	}
    }
    sgrep_free(as);
}
	
void delete_AC_scanner(struct ACScanner *ac) {
    /* FIXME: this leaks memory! */
    SGREPDATA(ac);
    delete_AC_state(sgrep,ac->root_state);
    sgrep_free(ac);
}


/* 
 * The AC automate search. 
 * (A dramatically simpler version, than which it used to be four years
 *  ago when it first saw the light of the day.
 *  It seems that i've actually gained some "programming experience"
 *  in these years :) 
 */
void ACsearch(struct ACScanner *scanner, const unsigned char *buf, 
	      int len, int start)
{
    struct OutputList *op;
    int ch;
    int i;
    struct ACState *s;

    s=scanner->s;
    for(i=0;i<len;i++) {
	ch=(scanner->ignore_case) ? toupper(buf[i]) : buf[i];
	while (s->gotos[ch]==NULL) {
	    assert(s->fail);
	    s=s->fail;
	}	
	s=s->gotos[ch];
	op=s->output_list;
	while(op!=NULL) {
	    scanner->sgrep->statistics.phrases++;
	    assert(op->phrase->regions!=NULL);
	    add_region( op->phrase->regions,
			i-(op->phrase->phrase->length-1)+start+1,
			i+start);
#ifdef DEBUG
	    printf("Found \"%s\" in file %s at %d  gc<-(%d,%d)\n",
		   op->phrase->phrase->s,
		   file_list->files[file_num].name,
		   file_pos-buf_end+buf_pos-(op->phrase->phrase->length),
		   i- ( op->phrase->phrase->length ) , i-1);
#endif
	    op=op->next;
	} while ( op!=NULL );
    }
    scanner->s=s;
}

SGMLScanner *init_sgml_scanner_common(SgrepData *sgrep) {    
    SGMLScanner *scanner;
    scanner=sgrep_new(SGMLScanner);
    scanner->sgrep=sgrep;
    scanner->state_stack_ptr=0;

    scanner->maintain_element_stack=1;
    scanner->top=NULL;
    scanner->element_list=NULL;

    scanner->name_start_chars=expand_character_list(sgrep,NameStartChars);
    scanner->name_chars=expand_character_list(sgrep,NameChars);
    if (sgrep->word_chars) {
	scanner->word_chars=expand_character_list(sgrep,sgrep->word_chars);
    } else {
	scanner->word_chars=expand_character_list(sgrep,WordChars);
    }
    scanner->parse_errors=0;

    scanner->type=sgrep->scanner_type;
    scanner->ignore_case=sgrep->ignore_case;
    scanner->include_system_entities=sgrep->include_system_entities;

    scanner->state=SGML_PCDATA;
    scanner->sgi[0]='s';
    scanner->egi[0]='e';
    scanner->aname[0]='a';
    scanner->aval[0]='v';
    scanner->word[0]='w';
    scanner->comment_word[0]='c';
    scanner->pi[0]='?';
    scanner->name[0]=0; /* Changed when needed */
    scanner->failed=0;
    return scanner;
}

void push_state(SGMLScanner *scanner, enum SGMLState state) {
    assert(scanner->state_stack_ptr<SGML_SCANNER_STACK_SIZE);
    scanner->state_stack[scanner->state_stack_ptr++]=state;
}

enum SGMLState pop_state(SGMLScanner *scanner) {
    assert(scanner->state_stack_ptr>0);
    return scanner->state_stack[--scanner->state_stack_ptr];
}


/* FIXME: needs hashing for speed */
void sgml_add_entry_to_gclist(SGMLScanner *state,
			      const char *phrase,int start, int end) {
    struct PHRASE_NODE *n;
    for(n=state->phrase_list;n!=NULL;n=n->next) {
	if (n->phrase->s[n->phrase->length-1]=='*') {
	    /* Wildcard */
	    if (strncmp(n->phrase->s,phrase,n->phrase->length-1)==0) {
		add_region(n->regions,start,end);
	    
	    }
	} else if (strcmp(n->phrase->s,phrase)==0) {
	    add_region(n->regions,start,end);
	}
    }
}

void sgml_add_entry_to_index(SGMLScanner *state,
			     const char *phrase,
			     int start, int end) {
    if (phrase[0]=='@') {
	add_region(state->element_list,start,end);
    } else {
	if (add_region_to_index((struct IndexWriterStruct *)state->data,
				phrase,start,end)==SGREP_ERROR) {
	    state->failed=1;
	}
    }
}

SGMLScanner *init_sgml_phrase_scanner(SgrepData *sgrep,
				      struct PHRASE_NODE *list) {    
    SGMLScanner *scanner;
    scanner=init_sgml_scanner_common(sgrep);
    scanner->phrase_list=list;
    scanner->entry=sgml_add_entry_to_gclist;
    scanner->data=NULL;
    return scanner;
}

void delete_sgml_scanner(SGMLScanner *s) {
    SgrepData *sgrep=s->sgrep;
    /* Empty the element stack if there is one */
    pop_elements_to(s,NULL);
    if (s->element_list) {
	delete_region_list(s->element_list);
    }
    sgrep_free(s->name_start_chars);
    sgrep_free(s->name_chars);
    sgrep_free(s->word_chars);
    sgrep_free(s);
}
    
SGMLScanner *init_sgml_index_scanner(SgrepData *sgrep,
				     struct IndexWriterStruct *writer) {    
    SGMLScanner *scanner;
    scanner=init_sgml_scanner_common(sgrep);
    scanner->phrase_list=NULL;
    scanner->element_list=new_region_list(sgrep);
    list_set_sorted(scanner->element_list,NOT_SORTED);
    scanner->element_list->nested=1;
    scanner->entry=sgml_add_entry_to_index;
    scanner->data=writer;
    return scanner;
}



#define SGML_ENTRY(QUERY,NAME,RAW_NAME,START,END) \
do { if (sgrep->sgml_debug) sgrep_error(sgrep,"%s(\"%s\"):%s:(%d,%d)\n",(QUERY),(NAME),(RAW_NAME),(START),(END)); \
if ((START)<=(END)) state->entry(state,(char *)(RAW_NAME),(START),(END)); } while (0)

void pop_elements_to(SGMLScanner *state,ElementStack *p) {
    ElementStack *q;
    SGREPDATA(state);
    assert(p==NULL || state->top);
    /* Element is in the stack */
    q=state->top;
    while(p!=q) {
	/* All elements in the stack not having an end tag
	 * are considered as empty. Sad but true */
	state->top=q->prev;
	SGML_ENTRY("elements","","@elements",q->start,q->end);
	/* fprintf(stderr,"<%s/>\n",q->gi); */
	sgrep_free(q->gi);
	sgrep_free(q);
	q=state->top;
    }
}

void sgml_found(SGMLScanner *state,int end_index) {
    int i;
    SGREPDATA(state);

    switch(state->state) {
    case SGML_WORD:
    case SGML_WORD_ENTITY:
	state->word[0]='w';
	if (state->ignore_case) {
	    for(i=1;state->word[i];i++) state->word[i]=tolower(state->word[i]);
	}
	SGML_ENTRY("word",state->word+1,state->word,state->words,end_index);
	break;

    case SGML_ENTITY:
	state->name[0]='&';
	SGML_ENTRY("entity",state->name+1,state->name,state->entitys,end_index);
	break;

    case SGML_CDATA_MARKED_SECTION_WORD:
	state->word[0]='w';
	if (state->ignore_case) {
	    for(i=1;state->word[i];i++) state->word[i]=tolower(state->word[i]);
	}
	SGML_ENTRY("word",state->word+1,state->word,state->words,end_index);
	break;

    case SGML_STAGC:
	if (state->type!=XML_SCANNER) 
	    for(i=1;state->sgi[i];i++) state->sgi[i]=toupper(state->sgi[i]);
	SGML_ENTRY("stag",state->sgi+1,state->sgi,state->sgis,end_index);
	if (state->maintain_element_stack) {
	    /* Push to element stack */
	    ElementStack *e=sgrep_new(ElementStack);
	    e->gi=sgrep_strdup((char *)state->sgi+1);
	    e->start=state->sgis;
	    e->end=end_index;
	    e->prev=state->top;
	    state->top=e;
	}
	break;

    case SGML_ATTVALUE:
    case SGML_ATTVALUE_DQUOTED:
    case SGML_ATTVALUE_SQUOTED:
	if (state->type!=XML_SCANNER) {
	    for(i=1;state->aname[i];i++) state->aname[i]=toupper(state->aname[i]);
	}
	SGML_ENTRY("attribute",state->aname+1,state->aname,state->anames,
		   end_index);
	SGML_ENTRY("attvalue",state->aval+1,state->aval,state->avals,
		   end_index);
	break;

    case SGML_W_ETAGC:
	if (state->type!=XML_SCANNER) 
	    for(i=1;state->egi[i];i++) state->egi[i]=toupper(state->egi[i]);
	SGML_ENTRY("etag",state->egi+1,state->egi,state->egis,end_index);
	if (state->maintain_element_stack) {
	    /* First check that the element is on the stack */
	    ElementStack *p=state->top;
	    while(p && strcmp((char*)state->egi+1,p->gi)!=0) p=p->prev;
	    if (p) {
		/* Take elements until p is in top */
		pop_elements_to(state,p);
		/* Pop p */
		state->top=p->prev;
		SGML_ENTRY("elements","","@elements",p->start,end_index);
		/* fprintf(stderr,"<%s>..</%s>\n",p->gi,p->gi);*/
		sgrep_free(p->gi);
		sgrep_free(p);
	    }
	}
	break;

    case SGML_COMMENT_WORD:
	if (state->ignore_case) {
	    for(i=1;state->comment_word[i];i++) 
		state->comment_word[i]=tolower(state->comment_word[i]);
	}
	SGML_ENTRY("comment_word",state->comment_word+1,state->comment_word,
		   state->comment_words,end_index);
	break;
	   
    case SGML_COMMENT_END2:
	SGML_ENTRY("comment",(const unsigned char *)"",
		   (const unsigned char *)"-",
		   state->comments,end_index);
	break;

    case SGML_PI:
    case SGML_PI_END:
	SGML_ENTRY("pi",state->pi+1,state->pi,state->pis,end_index);
	break;

    case SGML_CDATA_MARKED_SECTION_END2:
	SGML_ENTRY("cdata",(const unsigned char *)"",
		   (const unsigned char *)"[CDATA",
		   state->markeds,end_index);
	break;

    case SGML_DOCTYPE:
	state->name[0]='d';
	state->name[1]='n';
	if (state->type!=XML_SCANNER) {
	    for(i=2;state->name[i];i++) state->name[i]=toupper(state->name[i]);
	}
	SGML_ENTRY("doctype",state->name+2,state->name,state->doctypes,
		   end_index);
	/* Empty the element stack */
	pop_elements_to(state,NULL);
	break;

    case SGML_DOCTYPE_PUBLIC_ID:
	state->literal[1]='d';
	state->literal[2]='p';
	SGML_ENTRY("doctype_pid",state->literal+3,state->literal+1,
		   state->literals,end_index);
	break;
	
    case SGML_DOCTYPE_SYSTEM_ID:
	state->literal[1]='d';
	state->literal[2]='s';
	SGML_ENTRY("doctype_sid",state->literal+3,state->literal+1,
		   state->literals,end_index);
	break;

    case SGML_DOCTYPE_END:
	SGML_ENTRY("prologs",(const unsigned char *)"d",
		   (const unsigned char *)"d!",
		   state->doctype_declarations, end_index);
	break;

    case SGML_GENERAL_ENTITY_DEFINITION_DQUOTED:
    case SGML_GENERAL_ENTITY_DEFINITION_SQUOTED:
	state->name[0]='!';
	state->name[1]='e';
	state->name[2]='l';
	SGML_ENTRY("entity_literal",state->name+3,state->name,
		   state->literals,end_index);
	break;

    case SGML_ENTITY_DEFINITION_END:
	state->name[0]='!';
	state->name[1]='e';
	state->name[2]='d';
	SGML_ENTRY("entity_declaration",state->name+3,state->name,
		   state->internal_declarations,end_index);
	/* Check if this is an system entity reference we should include */
	if (state->entity_has_systemid && 
	    (!state->entity_is_ndata) &&
	    state->include_system_entities) {
	    if (!flist_exists(state->scan_buffer->file_list,
			      (char *)state->literal+3)) {
		if (flist_add_relative(state->scan_buffer->file_list,
				       state->scan_buffer->file_num,
				       (char *)state->literal+3)==SGREP_OK) {
		    sgrep_progress(sgrep,
				   "Including system entity '%s'\n",
				   state->literal+3);
		    /* fprintf(stderr,"File list size %d\n",
		       flist_files(state->scan_buffer->file_list)); */

		} else {
		    sgrep_progress(sgrep,"Cannot include system entity '%s'\n",
				   state->literal+3);
		}
	    }
	}
	break;

    case SGML_ENTITY_DEFINITION_PUBLIC_ID:
	state->literal[0]='!';
	state->literal[1]='e';
	state->literal[2]='p';
	SGML_ENTRY("entity_pid",state->literal+3,state->literal,
		   state->literals,end_index);
	break;

    case SGML_ENTITY_DEFINITION_SYSTEM_ID:
	state->literal[0]='!';
	state->literal[1]='e';
	state->literal[2]='s';
	SGML_ENTRY("entity_sid",state->literal+3,state->literal,
		   state->literals,end_index);
	break;

    case SGML_ENTITY_DEFINITION_NDATA_NAME:
	state->word[0]='!';
	state->word[1]='e';
	state->word[2]='n';
	if (state->type!=XML_SCANNER) {
	    for(i=3;state->word[i]!=0;i++) state->word[i]=toupper(state->word[i]);
	}
	SGML_ENTRY("entity_ndata",state->word+3,state->word,
		   state->words,end_index);
	break;

    case SGML_ELEMENT_TYPE_DECLARATION:
    case SGML_ATTLIST_DECLARATION:
    case SGML_NOTATION_DECLARATION:
	/* Not used */
	break;
	
    default:
	sgrep_error(sgrep,"SGML huh?\n");
	break;
    }
}
#undef DEBUG

int sgml_scan(SGMLScanner *scanner) {
#define NEXT_CH (ch=buf[i++])
#define state (scanner->state)
    int i;
    int ch;
    const int len=scanner->scan_buffer->len;
    const int start=scanner->scan_buffer->region_start;
    const unsigned char *buf=scanner->scan_buffer->map;
    
    i=0;
    NEXT_CH;
    while(i<=len) {
	if (scanner->failed) return SGREP_ERROR;

	switch (state) {

	case SGML_PCDATA:
	    switch (ch) {
	    case '<':
		state=SGML_STAGO;
		NEXT_CH;
		break;
	    default:
		if (scanner->word_chars[ch]) {
		    state=SGML_WORD;
		    scanner->wordi=2;
		    scanner->words=start+i-1;
		    scanner->word[1]=ch;
		} else if (ch=='&') {
		    push_state(scanner,SGML_PCDATA_ENTITY);
		    state=SGML_ENTITY_OPEN;
		}
		NEXT_CH;
	    }
	    break;	    

	case SGML_ENTITY_OPEN:
	    scanner->character_reference=0;
	    if (ch=='#') {
		scanner->entitys=start+i-2;
		state=SGML_CHARACTER_REFERENCE_OPEN;
		NEXT_CH;
	    } else if (scanner->name_start_chars[ch]) {
		scanner->entitys=start+i-2;
		scanner->name[1]=ch;
		scanner->namei=2;
		scanner->entitys=start+i-2;
		state=SGML_ENTITY;
		NEXT_CH;
	    } else {
		scanner->entitys=-1;
		state=pop_state(scanner);
	    }
	    break;

	case SGML_ENTITY:
	    if (scanner->name_chars[ch] && 
		scanner->namei<MAX_TERM_SIZE) {
		scanner->name[scanner->namei++]=ch;
		NEXT_CH;
	    } else {
		scanner->name[scanner->namei]=0;
		if (ch==';') {
		    sgml_found(scanner,start+i-1);
		    NEXT_CH;
		} else {
		    sgml_found(scanner,start+i-2);
		}
		if (scanner->namei<MAX_TERM_SIZE) {
		    scanner->name[scanner->namei++]=';';
		    scanner->name[scanner->namei]=0;
		}
		state=pop_state(scanner);
	    }
	    break;
	    
	case SGML_CHARACTER_REFERENCE_OPEN:
	    if (ch=='x') {
		state=SGML_HEX_CHARACTER_REFERENCE;
		NEXT_CH;
	    } else if (ch>='0' && ch<='9') {
		scanner->character_reference=ch-'0';
		state=SGML_DECIMAL_CHARACTER_REFERENCE;
		NEXT_CH;
	    } else {
		scanner->entitys=-1;
		scanner->character_reference=0;
		scanner->parse_errors++;
		state=pop_state(scanner);
	    }
	    break;
	    
	case SGML_DECIMAL_CHARACTER_REFERENCE:
	    if (ch==';') {
		state=SGML_CHARACTER_REFERENCE_CLOSE;
		NEXT_CH;
		break;
	    } else if (ch>='0' && ch<='9') {
		scanner->character_reference=
		    scanner->character_reference*10+ch-'0';
		NEXT_CH;
	    } else {		
		state=SGML_CHARACTER_REFERENCE_CLOSE;
	    }	    
	    break;

	case SGML_HEX_CHARACTER_REFERENCE:
	    if (ch==';') {
		state=SGML_CHARACTER_REFERENCE_CLOSE;
		NEXT_CH;
	    } else if (ch>='0' && ch<='9') {
		scanner->character_reference=
		    scanner->character_reference*16+ch-'0';
		NEXT_CH;
	    } else  if (toupper(ch)>='A' && toupper(ch)<='F') {
		scanner->character_reference=scanner->character_reference*16+
		    toupper(ch)-'A'+10;
		NEXT_CH;
	    } else {
		state=SGML_CHARACTER_REFERENCE_CLOSE;
	    }
	    break;
		
	case SGML_CHARACTER_REFERENCE_CLOSE:
	    sprintf((char *)scanner->name,"&#x%x;",
		    scanner->character_reference);
	    /* fprintf(stderr,"charref: %s\n",scanner->name); */
	    state=pop_state(scanner);
	    break;

	case SGML_PCDATA_ENTITY:
	    if (scanner->character_reference>0 && 
		scanner->character_reference<255 &&
		scanner->word_chars[scanner->character_reference]) {
		/* Entity was a character entity and was word character */
		scanner->words=scanner->entitys;
		scanner->word[1]=scanner->character_reference;
		scanner->wordi=2;
		state=SGML_WORD;
	    } else if (scanner->entitys>=0 && scanner->word_chars[ch]) {
		/* Handles the case when word starts with a some entity */
		scanner->words=scanner->entitys;
		scanner->wordi=scanner->namei+1;
		if (scanner->wordi>MAX_TERM_SIZE) {
		    scanner->wordi=MAX_TERM_SIZE;
		}
		strncpy((char *)scanner->word+1,(const char *)scanner->name,
			MAX_TERM_SIZE-1);
		scanner->word[scanner->wordi]=0;
		state=SGML_WORD;
	    } else {
		state=SGML_PCDATA;
	    }
	    break;


	case SGML_WORD_ENTITY:
	    if (scanner->character_reference>0 && 
		scanner->character_reference<255 &&
		scanner->word_chars[scanner->character_reference]) {
		/* Entity was a character entity and was word character */
		if (scanner->wordi<MAX_TERM_SIZE) {
		    scanner->word[scanner->wordi++]=scanner->character_reference;
		}
		state=SGML_WORD;
	    } else {
		/* Handles the case when word continues with some entity */
		if (scanner->entitys>=0 && 
		    (scanner->character_reference==0 ||
			scanner->character_reference>=255)) {
		    /* Use the located entity only if it vas valid */
		    strncpy((char *)scanner->word+scanner->wordi,
			    (const char *)scanner->name,MAX_TERM_SIZE-scanner->wordi);
		    scanner->wordi+=scanner->namei;
		    if (scanner->wordi>MAX_TERM_SIZE) {
			scanner->wordi=MAX_TERM_SIZE;
		    }
		    scanner->word[scanner->wordi]=0;
		    state=SGML_WORD;
		} else {
		    /* Entity ended the word: either in syntax error
		     * or non word char */
		    scanner->word[scanner->wordi]=0;
		    sgml_found(scanner,scanner->word_end);
		    state=SGML_PCDATA;
		}
	    }
	    break;
	    
	case SGML_WORD:
	    if (scanner->word_chars[ch] && ch!='<' &&
		scanner->wordi<MAX_TERM_SIZE) {
		scanner->word[scanner->wordi++]=ch;
		NEXT_CH;
	    } else if (ch=='&') {
		scanner->word_end=start+i-2;
		push_state(scanner,SGML_WORD_ENTITY);
		state=SGML_ENTITY_OPEN;
		NEXT_CH;
	    } else {
		scanner->word[scanner->wordi]=0;
		sgml_found(scanner,start+i-2);
		state=SGML_PCDATA;
	    }
	    break;

	case SGML_STAGO:
	    switch(ch) {
	    case '/':
		scanner->egii=1;
		scanner->egis=start+i-2;
		state=SGML_END_TAG;
		NEXT_CH;
		break;
	    case '!':
		state=SGML_DECLARATION_START;
		NEXT_CH;
		break;
	    case '?':
		scanner->pis=start+i-2;
		scanner->pii=1;
		state=SGML_PI_START;
		push_state(scanner,SGML_PCDATA);
		NEXT_CH;
		break;
	    default:
		if (scanner->name_start_chars[ch]) {
		    state=SGML_GI;
		    scanner->sgis=start+i-2;
		    scanner->sgii=2;
		    scanner->sgi[1]=ch;
		    NEXT_CH;
		} else {
		    state=SGML_PCDATA;
		}
		break;
	    }
	    break;

	case SGML_GI:
	    if (scanner->name_chars[ch] && scanner->sgii<MAX_TERM_SIZE) {
		scanner->sgi[scanner->sgii++]=ch;
		NEXT_CH;
	    } else {
		scanner->sgi[scanner->sgii]=0;
		state=SGML_W_ATTNAME;
	    }
	    break;

	case SGML_W_ATTNAME:
	    if (ch=='>') {
		state=SGML_STAGC;
	    } else if (scanner->name_start_chars[ch]) {
		scanner->anamei=2;
		scanner->anames=start+i-1;
		scanner->aname[1]=ch;
		state=SGML_ATTNAME;
		NEXT_CH;
	    } else {
		NEXT_CH;
	    }	    
	    break;

	case SGML_ATTNAME:
	    if (scanner->name_chars[ch] && 
		scanner->anamei<MAX_TERM_SIZE) {
		scanner->aname[scanner->anamei++]=ch;
		NEXT_CH;
	    } else {
		scanner->aname[scanner->anamei]=0;
		state=SGML_W_ATTEQUAL;
	    }
	    break;

	case SGML_W_ATTEQUAL:
	    switch(ch) {
	    case ' ': case '\t':  case '\n':  case '\r': case '\v':
		NEXT_CH;
		break;
	    case '=':
		state=SGML_W_ATTVALUE;
		NEXT_CH;
		break;
		
	    default:
		if (ch=='>' || scanner->name_start_chars[ch]) {
		    /* FIXME: handle attribute value with no name */
		    /* fprintf(stderr,"Attribute with no value\n"); */
		}
		if (ch=='>') {
		    state=SGML_STAGC;
		} else if (scanner->name_start_chars[ch]) {
		    state=SGML_W_ATTNAME;
		} else {
		    /* Parse error (unexpected character) */
		    scanner->parse_errors++;
		    state=SGML_W_ATTNAME;
		}
		break;
	    }
	    break;

	case SGML_W_ATTVALUE:
	    switch(ch) {
	    case ' ': case '\t':  case '\n':  case '\r': case '\v':
		NEXT_CH;
		break;
	    case '>':
		/* Parse error.. */
		scanner->parse_errors++;
		state=SGML_STAGC;
		break;
	    case '\"':
		state=SGML_ATTVALUE_DQUOTED;
		scanner->avals=start+i-1;
		scanner->avali=1;
		NEXT_CH;
		break;
	    case '\'':
		state=SGML_ATTVALUE_SQUOTED;
		scanner->avals=start+i;
		scanner->avali=1;
		NEXT_CH;
		break;
	    default:
		if (isgraph(ch)) {
		    state=SGML_ATTVALUE;
		    scanner->avals=start+i-1;
		    scanner->avali=2;
		    scanner->aval[1]=ch;
		    NEXT_CH;
		} else {
		    /* Parse error */
		    scanner->parse_errors++;
		    state=SGML_W_ATTNAME;
		    break;
		}
		break;
	    }
	    break;
	    
	case SGML_ATTVALUE:
	    if (ch=='>' || isspace(ch) || !isgraph(ch) || 
		scanner->avali>=MAX_TERM_SIZE-1) {
		scanner->aval[scanner->avali]=0;
		sgml_found(scanner,start+i-2);
		state=SGML_W_ATTNAME;
		break;
	    }
	    scanner->aval[scanner->avali++]=ch;
	    NEXT_CH;
	    break;

	case SGML_ATTVALUE_DQUOTED:
	    if (ch!='\"' && scanner->avali<=MAX_TERM_SIZE) {
		scanner->aval[scanner->avali++]=ch;
		NEXT_CH;
		break;
	    }
	    scanner->aval[scanner->avali]=0;
	    sgml_found(scanner,start+i-1);
	    state=SGML_W_ATTNAME;
	    NEXT_CH;
	    break;

	case SGML_ATTVALUE_SQUOTED:
	    if (ch!='\'' && scanner->avali<=MAX_TERM_SIZE) {
		scanner->aval[scanner->avali++]=ch;
		NEXT_CH;
		break;
	    }
	    scanner->aval[scanner->avali]=0;
	    sgml_found(scanner,start+i-1);
	    state=SGML_W_ATTNAME;
	    NEXT_CH;
	    break;
	    

	case SGML_STAGC:
	    sgml_found(scanner,start+i-1);
	    state=SGML_PCDATA;
	    NEXT_CH;
	    break;	    

	case SGML_END_TAG:
	    if (scanner->name_chars[ch] 
		&& scanner->egii<MAX_TERM_SIZE) {
		scanner->egi[scanner->egii++]=ch;
		NEXT_CH;
	    } else {
		state=SGML_W_ETAGC;
		scanner->egi[scanner->egii]=0;
	    }
	    break;
	    
	case SGML_W_ETAGC:
	    if (ch=='>') {
		sgml_found(scanner,start+i-1);
		state=SGML_PCDATA;
	    }
	    NEXT_CH;
	    break;

	case SGML_DECLARATION_START:
	    switch(ch) {
	    case '-':
		scanner->comments=start+i-3;
		state=SGML_COMMENT_START;
		push_state(scanner,SGML_PCDATA);
		NEXT_CH;
		break;
	    case '[':
		scanner->markeds=start+i-3;
		state=SGML_MARKED_SECTION_START;
		scanner->sgii=2;
		scanner->sgi[1]='[';
		NEXT_CH;
		break;
	    default:
		if (toupper(ch)=='D') {
		    state=SGML_DOCTYPE_DECLARATION;
		    scanner->doctype_declarations=start+i-3;
		    scanner->name[0]=toupper(ch);
		    scanner->namei=1;
		    NEXT_CH;
		} else {
		    /* Parse error */
		    scanner->parse_errors++;
		    state=SGML_PCDATA;
		}
	    }
	    break;

	case SGML_COMMENT_START:
	    if (ch=='-') {
		state=SGML_COMMENT;
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=pop_state(scanner);
	    }
	    break;
	    
	case SGML_COMMENT:
	    if (ch=='-') {
		state=SGML_COMMENT_END1;
	    } else if (scanner->word_chars[ch]) {
		    state=SGML_COMMENT_WORD;
		    scanner->comment_wordi=2;
		    scanner->comment_words=start+i-1;
		    scanner->comment_word[1]=ch;
	    }
	    NEXT_CH;
	    break;

	case SGML_COMMENT_WORD:
	    /* FIXME: This does not accept - as word char inside comments */
	    if (scanner->word_chars[ch] && ch!='-' &&
		scanner->comment_wordi<MAX_TERM_SIZE) {
		scanner->comment_word[scanner->comment_wordi++]=ch;
		NEXT_CH;
	    } else {
		scanner->comment_word[scanner->comment_wordi]=0;
		sgml_found(scanner,start+i-2);
		state=SGML_COMMENT;
	    }
	    break;

	case SGML_COMMENT_END1:
	    if (ch=='-') {
		state=SGML_COMMENT_END2;
		NEXT_CH;
	    } else {
		state=SGML_COMMENT;
	    }
	    break;

	    /* FIXME: generalize comment handling to situations
	     * like <!ATTLIST blah --sadfasdf-- blearhj --safsad--> */
	case SGML_COMMENT_END2:
	    if (ch=='>') {
		sgml_found(scanner,start+i-1);
		state=pop_state(scanner);
	    } else if (ch=='-') {
		state=SGML_COMMENT_START;
	    } else if (!isspace(ch)) {
		/* Since everything else except white space is
		 * a parse error */                 		   
		scanner->parse_errors++;
	    }
	    NEXT_CH;
	    break;

	case SGML_PI_START:
	    if ( ((scanner->pii==1 && scanner->name_start_chars[ch]) ||
		 (scanner->pii>1 && scanner->name_chars[ch])) &&
		 scanner->pii<MAX_TERM_SIZE
		) {
		scanner->pi[scanner->pii++]=ch;
		NEXT_CH;
	    } else {
		scanner->pi[scanner->pii]=0;
		state=SGML_PI;
	    }
	    break;

	case SGML_PI:
	    switch(ch) {
	    case '?':
		if (scanner->type==XML_SCANNER) state=SGML_PI_END;
		break;
	    case '>':
		if (scanner->type!=XML_SCANNER) {
		    sgml_found(scanner,start+i-1);
		    state=pop_state(scanner);
		}
		break;
	    }
	    NEXT_CH;
	    break;

	case SGML_PI_END:
	    switch(ch) {
	    case '?':
		break;
	    case '>':
		sgml_found(scanner,start+i-1);
		state=pop_state(scanner);
		break;
	    default:
		state=SGML_PI;
	    }
	    NEXT_CH;
	    break;

	case SGML_MARKED_SECTION_START:
	    /* Using gi also for marked section names. */
	    if (
		/* NOTE: entity references in marked section type are allowed
		 * to enable the SGML IGNORE and PCDATA entity tricks */
		(scanner->sgii==2 && 
		 (scanner->name_start_chars[ch] || ch=='&' || ch==';')) ||
		(scanner->sgii>2 &&
		 (scanner->name_chars[ch] || ch=='&' || ch==';')) ) {
		scanner->sgi[scanner->sgii++]=toupper(ch);
		NEXT_CH;
	    } else {
		scanner->sgi[scanner->sgii]=0;
		state=SGML_MARKED_SECTION_START2;
	    }
	    break;

	case SGML_MARKED_SECTION_START2:
	    if (isspace(ch)) {
		NEXT_CH;
	    } else if (ch=='['){
		if (strcmp((char *)scanner->sgi,"s[CDATA")==0) {
		    state=SGML_CDATA_MARKED_SECTION;
		} else {
		    /* Since extracting other than CDATA marked sections
		     * would need a stack this only reports the start of
		     * those sections. Maybe in version 3..*/
		    sgml_found(scanner,start+i-1);
		    state=SGML_PCDATA;
		}
		NEXT_CH;
	    } else {
		/* Parse error */
		scanner->parse_errors++;
		state=SGML_PCDATA;
	    }
	    break;
	    
	case SGML_CDATA_MARKED_SECTION:
	    if (ch==']') {
		state=SGML_CDATA_MARKED_SECTION_END1;
	    } else if (scanner->word_chars[ch]) {
		scanner->wordi=2;
		scanner->word[1]=ch;
		scanner->words=start+i-1;
		state=SGML_CDATA_MARKED_SECTION_WORD;
	    }
	    NEXT_CH;
	    break;

	case SGML_CDATA_MARKED_SECTION_WORD:
	    if (scanner->word_chars[ch] && ch!=']' &&
		scanner->wordi<MAX_TERM_SIZE) {
		scanner->word[scanner->wordi++]=ch;
		NEXT_CH;
	    } else {
		scanner->word[scanner->wordi]=0;
		sgml_found(scanner,start+i-2);
		state=SGML_CDATA_MARKED_SECTION;
	    }
	    break;

	case SGML_CDATA_MARKED_SECTION_END1:
	    if (ch==']') {
		state=SGML_CDATA_MARKED_SECTION_END2;
		NEXT_CH;
	    } else {
		state=SGML_CDATA_MARKED_SECTION;
	    }
	    break;

	case SGML_CDATA_MARKED_SECTION_END2:
	    if (ch=='>') {
		sgml_found(scanner,start+i-1);
		NEXT_CH;
		state=SGML_PCDATA;
	    } else if (ch==']') {
		NEXT_CH;
	    } else {
		state=SGML_CDATA_MARKED_SECTION;
	    }
	    break;

	case SGML_DOCTYPE_DECLARATION: 
	    if (scanner->namei<7 && toupper(ch)==("DOCTYPE")[scanner->namei]) {
		scanner->namei++;
		NEXT_CH;
	    } else if (scanner->namei==7 && isspace(ch)) {
		scanner->namei=2;
		state=SGML_DOCTYPE;
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=SGML_PCDATA;
	    }
	    break;

	case SGML_DOCTYPE:
	    if (scanner->namei==2) {
		if (scanner->name_start_chars[ch]) {
		    scanner->name[scanner->namei++]=ch;
		    scanner->doctypes=start+i-1;
		    NEXT_CH;
		} else if (!isspace(ch)) {
		    scanner->parse_errors++;
		    state=SGML_PCDATA;
		} else {
		    NEXT_CH;
		}
	    } else {
		if (scanner->name_chars[ch] &&
		    scanner->namei<MAX_TERM_SIZE) {
		    scanner->name[scanner->namei++]=ch;
		    NEXT_CH;
		} else {
		    scanner->name[scanner->namei]=0;
		    sgml_found(scanner,start+i-2);
		    state=SGML_DOCTYPE_EXTERNAL;
		}
	    }
	    break;

	case SGML_DOCTYPE_EXTERNAL:
	    switch(ch) {
	    case '[':
		state=SGML_DOCTYPE_INTERNAL;
		NEXT_CH;
		break;
	    case 'P': case 'p':
		/* _P_ublic */
		scanner->publici=1;
		state=SGML_DOCTYPE_PUBLIC;
		NEXT_CH;
		break;
	    case 'S': case 's':
		scanner->systemi=1;
		state=SGML_DOCTYPE_SYSTEM;
		NEXT_CH;
		break;
	    case '>':
		state=SGML_DOCTYPE_END;
		break;
	    default:
		if (!isspace(ch)) {
		    scanner->parse_errors++;
		    state=SGML_PCDATA;
		}
		NEXT_CH;
	    }
	    break;
		
	case SGML_DOCTYPE_PUBLIC:
	    if (("PUBLIC")[scanner->publici]==toupper(ch)) {
		scanner->publici++;
		if (scanner->publici==6) {
		    state=SGML_DOCTYPE_PUBLIC_ID_START;
		}
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=SGML_PCDATA;
	    }
	    break;

	case SGML_DOCTYPE_SYSTEM:
	    if (("SYSTEM")[scanner->systemi]==toupper(ch)) {
		scanner->systemi++;
		if (scanner->systemi==6) {
		    state=SGML_DOCTYPE_SYSTEM_ID_START;
		}
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=SGML_PCDATA;
	    }
	    break;

	case SGML_DOCTYPE_PUBLIC_ID_START:
	    if (ch=='"' || ch=='\'') {
		state=SGML_LITERAL_START;
		push_state(scanner,SGML_DOCTYPE_PUBLIC_ID);
	    } else if (isspace(ch)) {
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=SGML_PCDATA;	
	    }
	    break;

	case SGML_WAITING_LITERAL:
	    if (isspace(ch)) {
		NEXT_CH;
	    } else {
		if (ch=='"' || ch=='\'') {
		    state=SGML_LITERAL_START;
		} else {
		    scanner->parse_errors++;
		    scanner->literals=-1;
		    state=pop_state(scanner);
		}
	    }
	    break;

	case SGML_LITERAL_START:
	    scanner->literals=start+i;
	    scanner->literali=3;
	    state=(ch=='"') ? SGML_LITERAL_DQUOTED : SGML_LITERAL_SQUOTED;
	    NEXT_CH;
	    break;

	case SGML_LITERAL_DQUOTED:
	    if (ch!='"' && scanner->literali<MAX_TERM_SIZE) {
		scanner->literal[scanner->literali++]=ch;
	    } else {
		scanner->literal[scanner->literali]=0;
		state=pop_state(scanner);
	    }
	    NEXT_CH;
	    break;

	case SGML_LITERAL_SQUOTED:
	    if (ch!='\'' && scanner->literali<MAX_TERM_SIZE) {
		scanner->literal[scanner->literali++]=ch;
	    } else {
		scanner->literal[scanner->literali]=0;
		state=pop_state(scanner);
	    }
	    NEXT_CH;
	    break;
	    
	case SGML_DOCTYPE_PUBLIC_ID:
	    sgml_found(scanner,start+i-3);
	    state=SGML_DOCTYPE_SYSTEM_ID_START;
	    break;

	case SGML_DOCTYPE_SYSTEM_ID_START:
	    if (ch=='[') {
		state=SGML_DOCTYPE_INTERNAL;
	    } else if (ch=='"' || ch=='\'') {
		state=SGML_LITERAL_START;
		push_state(scanner,SGML_DOCTYPE_SYSTEM_ID);
	    } else if (isspace(ch)) {
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=SGML_PCDATA;	
	    }
	    break;

	case SGML_DOCTYPE_SYSTEM_ID:
	    sgml_found(scanner,start+i-3);
	    state=SGML_DOCTYPE_INTERNAL_START;
	    break;
	    
	case SGML_DOCTYPE_INTERNAL_START:
	    switch(ch) {
	    case '[':
		state=SGML_DOCTYPE_INTERNAL;
		NEXT_CH;
		break;
	    case '>':
		state=SGML_DOCTYPE_END;
		break;
	    default:
		if (isspace(ch)) {
		    NEXT_CH;
		} else {
		    scanner->parse_errors++;
		    state=SGML_PCDATA;
		}
		break;
	    }
	    break;		    

	case SGML_DOCTYPE_INTERNAL:
	    switch(ch) {
	    case '<':
		state=SGML_INTERNAL_DECLARATION_START1;
		break;
	    case '%':
		state=SGML_PEREFERENCE;
		break;
	    case ']':
		state=SGML_DOCTYPE_END;
		break;
	    default:
		if (!isspace(ch)) {
		scanner->parse_errors++;
		}
		break;
	    }
	    NEXT_CH;
	    break;

	case SGML_PEREFERENCE:
	    if (!scanner->name_chars[ch]) {
		if (ch!=';' && !isspace(ch)) {
		    scanner->parse_errors++;
		}
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    NEXT_CH;
	    break;
		
	case SGML_INTERNAL_DECLARATION_START1:
	    switch (ch) {
	    case '!':
		state=SGML_INTERNAL_DECLARATION_START2;
		NEXT_CH;
		break;
	    case '?':
		scanner->pis=start+i-2;
		scanner->pii=1;
		state=SGML_PI_START;
		push_state(scanner,SGML_DOCTYPE_INTERNAL);
		NEXT_CH;
		break;
	    default:
		scanner->parse_errors++;
		state=SGML_DOCTYPE_INTERNAL;
		break;
	    }
	    break;
	    
	case SGML_INTERNAL_DECLARATION_START2:
	    if (ch=='-') {
		scanner->comments=start+i-3;
		push_state(scanner,SGML_DOCTYPE_INTERNAL);
		state=SGML_COMMENT_START;
		NEXT_CH;
	    } else if (isalpha(ch)) {
  		state=SGML_INTERNAL_DECLARATION_NAME;
		scanner->internal_declarations=start+i-3;
		scanner->name[0]=toupper(ch);
		scanner->namei=1;
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    break;
		
	case SGML_INTERNAL_DECLARATION_NAME: 
	    if (isalpha(ch) &&
		scanner->namei<10) {
		scanner->name[scanner->namei++]=toupper(ch);
		NEXT_CH;
	    } else if (isspace(ch)) {
		scanner->name[scanner->namei]=0;
		if (strcmp((char *)scanner->name,"ENTITY")==0) {
		    scanner->entity_has_systemid=0;
		    scanner->entity_is_ndata=0;
		    state=SGML_ENTITY_DECLARATION;
		} else if (strcmp((char *)scanner->name,"ELEMENT")==0) {
		    state=SGML_ELEMENT_TYPE_DECLARATION;
		} else if (strcmp((char *)scanner->name,"NOTATION")==0) {
		    state=SGML_NOTATION_DECLARATION;
		} else if (strcmp((char *)scanner->name,"ATTLIST")==0) {
		    state=SGML_ATTLIST_DECLARATION;
		} else {
		    scanner->parse_errors++;
		    state=SGML_DOCTYPE_INTERNAL;
		}
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    break;

	case SGML_ENTITY_DECLARATION:
	    if (ch=='%') {
		/* Parameter entities are handled just like any other
		 * entities */
		NEXT_CH;
	    } else if (scanner->name_start_chars[ch]) {
		state=SGML_ENTITY_DECLARATION_NAME;
		scanner->name[3]=ch;
		scanner->namei=4;
		NEXT_CH;
	    } else if (isspace(ch)) {
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    break;

	case SGML_ENTITY_DECLARATION_NAME:
	    if (scanner->name_chars[ch] && scanner->namei<MAX_TERM_SIZE) {
		scanner->name[scanner->namei++]=ch;
		NEXT_CH;
	    } else {
		scanner->name[scanner->namei]=0;
		state=SGML_ENTITY_DEFINITION;
	    }
	    break;
	    
	case SGML_ENTITY_DEFINITION:
	    if (ch=='"') {
		state=SGML_GENERAL_ENTITY_DEFINITION_DQUOTED;
		scanner->literals=start+i;
		NEXT_CH;
	    } else if (ch=='\'') {
		state=SGML_GENERAL_ENTITY_DEFINITION_SQUOTED;
		scanner->literals=start+i;
		NEXT_CH;
	    } else if (isspace(ch)) {
		NEXT_CH;
	    } else if (isalpha(ch)) {
		push_state(scanner,SGML_ENTITY_DEFINITION_TYPE);
		scanner->wordi=0;
		state=SGML_RESERVED_WORD;
	    } else {
		scanner->parse_errors++;
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    break;

	case SGML_RESERVED_WORD:
	    if (isalpha(ch) && scanner->wordi<MAX_TERM_SIZE) {
	       scanner->word[scanner->wordi++]=toupper(ch);
	       NEXT_CH;
	    } else {
		scanner->word[scanner->wordi]=0;
		state=pop_state(scanner);
	    }
	    break;

	case SGML_ENTITY_DEFINITION_TYPE:
	    if (strcmp((char *)scanner->word,"SYSTEM")==0) {
		push_state(scanner,SGML_ENTITY_DEFINITION_SYSTEM_ID);
		state=SGML_WAITING_LITERAL;
	    } else if (strcmp((char *)scanner->word,"PUBLIC")==0) {
		push_state(scanner,SGML_ENTITY_DEFINITION_PUBLIC_ID);
		state=SGML_WAITING_LITERAL;
	    } else if (strcmp((char *)scanner->word,"CDATA")==0 ||
		       strcmp((char *)scanner->word,"SDATA")==0) {
		state=SGML_LITERAL_ENTITY;
	    } else {
		scanner->parse_errors++;
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    break;
		    
	case SGML_LITERAL_ENTITY:
	    if (ch=='"') {
		state=SGML_GENERAL_ENTITY_DEFINITION_DQUOTED;
		scanner->literals=start+i;
		NEXT_CH;
	    } else if (ch=='\'') {
		state=SGML_GENERAL_ENTITY_DEFINITION_SQUOTED;
		scanner->literals=start+i;
		NEXT_CH;
	    } else if (isspace(ch)) {
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
	    }
	    break;
		
		
	case SGML_GENERAL_ENTITY_DEFINITION_DQUOTED:
	    if (ch=='"') {
		sgml_found(scanner,start+i-2);
		state=SGML_ENTITY_DEFINITION_END;
	    }
	    NEXT_CH;
	    break;
		    
	case SGML_GENERAL_ENTITY_DEFINITION_SQUOTED:
	    if (ch=='\'') {	
		sgml_found(scanner,start+i-2);
		state=SGML_ENTITY_DEFINITION_END;
	    }
	    NEXT_CH;
	    break;
	
	case SGML_ENTITY_DEFINITION_END:
	    if (ch=='>') {
		sgml_found(scanner,start+i-1);
		state=SGML_DOCTYPE_INTERNAL;
		NEXT_CH;
		break;
	    } else if (isspace(ch)) {
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    break;
	    
	case SGML_ENTITY_DEFINITION_PUBLIC_ID:
	    if (scanner->literals>0) {
		sgml_found(scanner,start+i-3);
		state=SGML_WAITING_ENTITY_DEFINITION_SYSTEM_ID;
	    } else {
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    break;
	    
	case SGML_WAITING_ENTITY_DEFINITION_SYSTEM_ID:
	    switch(ch) {
	    case '"':
	    case '\'':
	        push_state(scanner,SGML_ENTITY_DEFINITION_SYSTEM_ID);
		state=SGML_LITERAL_START;
		break;
	    default:
		if (isspace(ch)) {
		    NEXT_CH;
		} else {
		    state=SGML_ENTITY_DEFINITION_NDATA;
		}
	    }
	    break;
	    
	case SGML_ENTITY_DEFINITION_SYSTEM_ID:
	    if (scanner->literals>0) {
		sgml_found(scanner,start+i-3);
		scanner->entity_has_systemid=1;
		state=SGML_ENTITY_DEFINITION_NDATA;
	    }
	    break;
	    
	case SGML_ENTITY_DEFINITION_NDATA:
	    if (ch=='>') {
		state=SGML_ENTITY_DEFINITION_END;
	    } else if (isalpha(ch)) {
		scanner->wordi=0;
		push_state(scanner,SGML_ENTITY_DEFINITION_NDATA2);
		state=SGML_RESERVED_WORD;
	    } else if (isspace(ch)) {
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    break;

	case SGML_ENTITY_DEFINITION_NDATA2:
	    if (strcmp((char *)scanner->word,"NDATA")==0) {
		scanner->entity_is_ndata=1;
		if (isspace(ch)) {
		    NEXT_CH;
		} else if (scanner->name_start_chars[ch]){
		    scanner->wordi=3;
		    scanner->words=start+i-1;
		    state=SGML_ENTITY_DEFINITION_NDATA_NAME;
		} else {
		    scanner->parse_errors++;
		    state=SGML_DOCTYPE_INTERNAL;
		}
	    } else {
		scanner->parse_errors++;
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    break;

	case SGML_ENTITY_DEFINITION_NDATA_NAME:
	    if (scanner->name_chars[ch] && scanner->wordi<MAX_TERM_SIZE) {
		scanner->word[scanner->wordi++]=ch;
		NEXT_CH;
	    } else {
		scanner->word[scanner->wordi]=0;
		sgml_found(scanner,start+i-2);		
		state=SGML_ENTITY_DEFINITION_END;
	    }
	    break;

	    /* FIXME: could be useful */
	case SGML_ELEMENT_TYPE_DECLARATION:
	    if (ch=='>') {
		sgml_found(scanner,start+1);
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    NEXT_CH;
	    break;

	    /* FIXME: could be useful */
	case SGML_ATTLIST_DECLARATION:
	    if (ch=='>') {
		sgml_found(scanner,start+1);
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    NEXT_CH;
	    break;
	    
	    /* FIXME: could be useful */
	case SGML_NOTATION_DECLARATION:
	    if (ch=='>') {
		sgml_found(scanner,start+1);
		state=SGML_DOCTYPE_INTERNAL;
	    }
	    NEXT_CH;
	    break;
	

	case SGML_DOCTYPE_END:
	    if (ch=='>') {
		sgml_found(scanner,start+i);
		state=SGML_PCDATA;
		NEXT_CH;
	    } else if (isspace(ch)) {
		NEXT_CH;
	    } else {
		scanner->parse_errors++;
		state=SGML_PCDATA;
	    }
	    break;

	default:
	    sgrep_error(scanner->sgrep,
			"SGML-scanner in unimplemented state. Switching to PCDATA\n");
	    state=SGML_PCDATA;
	    break;
	}
	

    }
    return SGREP_OK;
#undef state
#undef NEXT_CH
}

int search(SgrepData *sgrep,struct PHRASE_NODE *phrase_list, FileList *files, 
	   int f_file, int l_file) {
    int sgml_phrases;
    int regex_phrases;
    int ac_phrases;
    int file_phrases;
    int e=SGREP_OK;

    /* If there is no phrases, there is no point to do searching */
    if (phrase_list==NULL) {
	sgrep_progress(sgrep,"No phrases. Skipping search\n");
	return SGREP_OK;
    }
    if (sgrep->index_file==NULL) {
	struct ScanBuffer *sb=NULL;
	struct ACScanner *acs=NULL;
	SGMLScanner *sgmls=NULL;
	struct PHRASE_NODE *j=NULL;

	file_phrases=ac_phrases=sgml_phrases=regex_phrases=0;

	/* We have to create empty gc lists for phrases */
	for (j=phrase_list;j!=NULL;j=j->next)
	{
	    assert(j->regions==NULL);
	    j->regions=new_region_list(sgrep);
	    if (j->phrase->s[0]=='@' ||
		j->phrase->s[0]=='*') {
		list_set_sorted(j->regions,NOT_SORTED);
		j->regions->nested=1;
	    }
	    
	    switch (j->phrase->s[0]) {
	    case 'n':
		ac_phrases++;
		break;
	    case 'r':
		regex_phrases++;
		break;
	    case 'f':
		file_phrases++;
		break;
	    case '#':
		/* Input independent phrases */
		break;
	    default:
		sgml_phrases++;
	    }	
	}

	/* Initialization */
	sb=new_scan_buffer(sgrep,files);
	reset_scan_buffer(sb,f_file,l_file);
	if (ac_phrases) {
	    acs=init_AC_search(sgrep,phrase_list);
	}
	if (sgml_phrases) {
	    sgmls=init_sgml_phrase_scanner(sgrep,phrase_list);
	}
	
	/* Main scanning loop, only if there is something to scan */
	if (acs || sgmls) while((e=next_scan_buffer(sb))>0) {
	    if (flist_files(files)>1) {
		sgrep_progress(sgrep,"Scanning %d/%d files %d/%dK (%d%%)\r",
			       sb->file_num,flist_files(files),
			       sb->region_start/1024,flist_total(files)/1024,
			       sb->region_start/(flist_total(files)/100+1));
	    } else {
		sgrep_progress(sgrep,"Scanning file '%s' %d/%dK (%d%%)\r",
			       flist_name(sb->file_list,sb->file_num),
			       sb->region_start/1024,flist_total(files)/1024,
			       sb->region_start/(flist_total(files)/100+1));
	    }		    
	    if (sgrep->progress_callback) {
		sgrep->progress_callback(sgrep->progress_data,
					 sb->file_num,flist_files(files),
					 sb->region_start,flist_total(files)); 
	    }			     	    
	    if (ac_phrases) {
		ACsearch(acs,sb->map,sb->len,sb->region_start);
	    }
	    if (sgml_phrases) {
		sgmls->scan_buffer=sb;
		sgml_scan(sgmls);
	    }
	}
	if (acs || sgmls) {
	    if (l_file==flist_files(files)-1) {
		sgrep_progress(sgrep,"\n");
	    } else {
		sgrep_progress(sgrep,
			       "                                            \r");
	    }
	}

#if 0
	/* FIXME: think this over */
	if (sgmls && sgmls->parse_errors>0) {
	    fprintf(stderr,"There was %d SGML parse errors\n",
		    sgmls->parse_errors);
	}
#endif
	/* Clean up scanners */
	delete_scan_buffer(sb);
	if (sgmls) {
	    delete_sgml_scanner(sgmls);
	}	
	if (acs) delete_AC_scanner(acs);

	/* Now handle the phrases, whose contents we know only after
	 * scanning or which are independent of scanning */
	/* FIXME and include-entities will break this */
	for(j=phrase_list;j!=NULL; j=j->next) {

	    switch(j->phrase->s[0]) {

	    case '#':
		if (strcmp(j->phrase->s,"#start")==0) {
		    int start=flist_start(files,f_file);
		    add_region(j->regions,start,start);
		} else if (strcmp(j->phrase->s,"#end")==0) {
		    int last=flist_start(files,l_file)+
			flist_length(files,l_file)-1;
		    add_region(j->regions,last,last);
		} else {
		    sgrep_error(sgrep,"Don't know how to handle phrase %s\n",
				j->phrase->s);
		}
		break;
		
	    case 'f': {
		int f;	       
		for(f=f_file;f<=l_file;f++) {
		    /* Check for filename */
		    if (j->phrase->s[j->phrase->length-1]=='*') {
			/* Wildcard */
			if (strncmp((char *)j->phrase->s+1,flist_name(files,f),
				    j->phrase->length-2)==0 &&
			    flist_length(files,f)>0) {
			    add_region(j->regions,
				       flist_start(files,f),
				       flist_start(files,f)+
				       flist_length(files,f)-1);
			}
		    } else if (strcmp((char *)j->phrase->s+1,
				      flist_name(files,f))==0 &&
			       flist_length(files,f)>0) {
			add_region(j->regions,
				   flist_start(files,f),
				   flist_start(files,f)+
				   flist_length(files,f)-1);
		    }
		}
	    }
	    break;
	    }
	}
    } else {
	sgrep_progress(sgrep,"Using lazy index file mode\n"); 
	e=SGREP_OK;
    }
    return (e==SGREP_ERROR) ? SGREP_ERROR:SGREP_OK;
}

/* FIXME: merge this better with search() */
int index_search(SgrepData *sgrep,struct IndexWriterStruct *writer,
		  FileList *files) {
    struct ScanBuffer *sb;
    SGMLScanner *sgmls;
    sb=new_scan_buffer(sgrep,files);
    sgmls=init_sgml_index_scanner(sgrep,writer);
    while(next_scan_buffer(sb)>0) {
	sgrep_progress(sgrep,"Indexing %d/%d files %d/%dK (%d%%)\r",
		      sb->file_num,flist_files(files),
		      sb->region_start/1024,flist_total(files)/1024,
		      sb->region_start/(flist_total(files)/100+1));	
	if (sgrep->progress_callback) {
	    sgrep->progress_callback(sgrep->progress_data,
				     sb->file_num,flist_files(files),
				     sb->region_start,flist_total(files)); 
	}
	sgmls->scan_buffer=sb;
	if (sgml_scan(sgmls)==SGREP_ERROR) {
	    sgrep_progress(sgrep,"\n");
	    delete_scan_buffer(sb);
	    delete_sgml_scanner(sgmls);
	    return SGREP_ERROR;
	}
    }
    sgrep_progress(sgrep,"\n");
    pop_elements_to(sgmls,NULL);
    if (sgmls->element_list) {
	ListIterator l;
	Region r;
	sgrep_progress(sgrep,"Adding element list to index\n");
	start_region_search(sgmls->element_list,&l);
	get_region(&l,&r);
	while(r.start!=-1) {
	    add_region_to_index(writer,"@elements",r.start,r.end);	    
	    get_region(&l,&r);
	}
    }
	
    delete_scan_buffer(sb);
    delete_sgml_scanner(sgmls);
    return SGREP_OK;
}


#ifdef DEBUG
/*
 * These are (were) used for debugging the creation of ac-automate
 */
void show_ACtree()
{
	printf("-------------------\n AC-tree\n------------------\n");
	show_states(0,root_state);
	show_outputs(root_state); 
}

void show_states(int depth,struct State_Data *s)
{
	int end=0;
	int i,j;
	
	for(i=0;i<256;i++)
	{
		if (s->gotos[i]!=NULL && 
		   !( s==root_state && s->gotos[i]==root_state ) )
		{
			if (end>0)
			for (j=0;j<depth;j++) printf("                ");
			printf("%2d:f=%2d:",s->state_num,s->fail->state_num);
			end++;
			printf("%c->%2d , ",i,s->gotos[i]->state_num);
			show_states(depth+1,s->gotos[i]);
		}
	}
	if (!end)
	{
		printf("%2d:f=%2d <---\n",s->state_num,s->fail->state_num);
	}
}

void show_outputs(struct State_Data *s)
{
	int i;
	struct OutPut *op;
	
	op=&s->out_list;
	if (op->phrase!=NULL)
	{
		printf("state %d:",s->state_num);
		do {
			printf(" %s",op->phrase->phrase->s);
			op=op->next;
		} while ( op!=NULL );
		printf("\n");
	}
	
	for (i=0;i<256;i++)
	{
		if (s->gotos[i]!=NULL && s->gotos[i]!=root_state )
			show_outputs(s->gotos[i]);
	}
}		
#endif
